package io.tswf.groovy.bettergroovy.transforms.extensions


import org.codehaus.groovy.ast.ClassNode

import java.lang.reflect.Modifier
import java.util.concurrent.ConcurrentLinkedDeque
import java.util.concurrent.CopyOnWriteArrayList

class ExtensionMethodRegistry {
    static Queue<ClassNode> foundInstanceExtensionClassNodes = new ConcurrentLinkedDeque<>()
    static Queue<ClassNode> foundStaticExtensionClassNodes = new ConcurrentLinkedDeque<>()
    private static volatile List<ExtensionMethodInfo> _instanceExtensionMethods
    private static volatile List<ExtensionMethodInfo> _staticExtensionMethods
    private static final Object _INSTANCE_EXTENSIONS_LOCK = new Object()
    private static final Object _STATIC_EXTENSIONS_LOCK = new Object()

    static List<ExtensionMethodInfo> getAllInstanceExtensions() {
        if (_instanceExtensionMethods == null) {
            synchronized (_INSTANCE_EXTENSIONS_LOCK) {
                if (_instanceExtensionMethods == null) {
                    _instanceExtensionMethods = new CopyOnWriteArrayList<>(loadExtensionsFromNodes(foundInstanceExtensionClassNodes))
                }
            }
        }

        return _instanceExtensionMethods
    }

    static List<ExtensionMethodInfo> getAllStaticExtensions() {
        if (_staticExtensionMethods == null) {
            synchronized (_STATIC_EXTENSIONS_LOCK) {
                if (_staticExtensionMethods == null) {
                    _staticExtensionMethods = new CopyOnWriteArrayList(loadExtensionsFromNodes(foundStaticExtensionClassNodes))
                }
            }
        }

        return _staticExtensionMethods
    }

    static Collection<ExtensionMethodInfo> loadExtensionsFromNodes(Collection<ClassNode> nodes) {
        def result = new ArrayList<ExtensionMethodInfo>()

        nodes.stream().distinctBy {it.name}.forEach { cn ->
            def methods = parseMethods(cn)
            result.addAll(methods)
        }

        return result
    }

    static List<ExtensionMethodInfo> parseMethods(ClassNode node) {
        def classIsExtensionsContainer = node.annotations.any {it.classNode.name == ExtensionMethodsContainer.class.name}

        if (!classIsExtensionsContainer) {
            return new ArrayList<ExtensionMethodInfo>()
        }
        
        return node.methods.stream()
            .filter { Modifier.isStatic(it.modifiers) && Modifier.isPublic(it.modifiers)}
            .filter {it.parameters.isNotNullOrEmpty()}
            .map {
                def extensionInfo = new ExtensionMethodInfo(
                    method: it,
                    targetType: it.parameters[0].type
                )

                return extensionInfo
            }
            .toList()
    }
}