package io.tswf.groovy.bettergroovy.transforms.extensions

import org.codehaus.groovy.ast.*
import org.codehaus.groovy.ast.expr.*
import org.codehaus.groovy.classgen.asm.sc.StaticTypesTypeChooser
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.transform.sc.StaticCompilationMetadataKeys

import java.lang.annotation.ElementType
import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy
import java.lang.annotation.Target

import static org.codehaus.groovy.ast.tools.GeneralUtils.args
import static org.codehaus.groovy.ast.tools.GeneralUtils.nullX

@GroovyASTTransformationClass(classes = SameSourceSetExtensionsTransformation)
@Target(ElementType.TYPE)
@Retention(RetentionPolicy.SOURCE)
@interface SameSourceSetExtensionMethods {

}

@GroovyASTTransformation(phase = CompilePhase.INSTRUCTION_SELECTION)
class SameSourceSetExtensionsTransformation extends MethodCallTransformation {

    @Override
    protected GroovyCodeVisitor getTransformer(ASTNode[] nodes, SourceUnit sourceUnit) {
        def target = nodes[1]
        return new ExtensionsMethodInvocationTransformer(
                sourceUnit: sourceUnit,
                classNode: target instanceof ClassNode ? target as ClassNode : null,
                instanceExtensions: ExtensionMethodRegistry.getAllInstanceExtensions(),
                staticExtensions: ExtensionMethodRegistry.getAllStaticExtensions()
        )
    }
}

class ExtensionsMethodInvocationTransformer extends ClassCodeExpressionTransformer {
    private final StaticTypesTypeChooser typeChooser = new StaticTypesTypeChooser()

    SourceUnit sourceUnit
    ClassNode classNode
    List<ExtensionMethodInfo> instanceExtensions
    List<ExtensionMethodInfo> staticExtensions

    @Override
    Expression transform(Expression expr) {
        if (classNode != null && hasAnyExtensionsHere()) {
            if (expr instanceof MethodCall) {
                return transformMethodCall(expr)
            }
        }

        return super.transform(expr)
    }

    boolean hasAnyExtensionsHere() {
        return instanceExtensions.isNotNullOrEmpty() || staticExtensions.isNotNullOrEmpty()
    }

    Expression transformMethodCall(Expression callExpression) {
        def dynamic = null

        if (callExpression instanceof MethodCallExpression) {
            dynamic = callExpression.objectExpression.getNodeMetaData(StaticCompilationMetadataKeys.RECEIVER_OF_DYNAMIC_PROPERTY)
        }

        if (dynamic == null) {
            def calledMethodOwnerClass = (ClassNode) null
            def callArgs = (Expression) null
            def calledMethodName = (String) null
            def isCallStatic = false
            def receiverExpression = (ASTNode) null

            if (callExpression instanceof MethodCallExpression) {
                calledMethodOwnerClass = typeChooser.resolveType(callExpression.objectExpression, classNode)
                callArgs = callExpression.arguments
                calledMethodName = callExpression.methodAsString
                isCallStatic = calledMethodOwnerClass.name == Class.class.getName() && calledMethodOwnerClass?.genericsTypes?.length == 1

                if (!isCallStatic) {
                    receiverExpression = callExpression.receiver
                }
            } else if (callExpression instanceof StaticMethodCallExpression) {
                calledMethodOwnerClass = callExpression.ownerType
                callArgs = callExpression.arguments
                calledMethodName = callExpression.methodAsString
                isCallStatic = true
            }

            calledMethodOwnerClass = unpackStaticClassReciever(calledMethodOwnerClass)

            if (calledMethodOwnerClass != null && callArgs instanceof ArgumentListExpression && (receiverExpression instanceof Expression || isCallStatic)) {
                def availableExtensions = isCallStatic ? getOwnerMatchingStaticExtensions(calledMethodOwnerClass) : getOwnerMatchingInstanceExtensions(calledMethodOwnerClass)

                if (availableExtensions.isNotNullOrEmpty()) {
                    def callArgsList = callArgs.toList()
                    def callArgsTypes = callArgsList.map {
                        it.isNullX() ? null : typeChooser.resolveType(it, classNode)
                    }

                    def targetMethod = calledMethodOwnerClass.getDeclaredMethods(calledMethodName)
                            .stream()
                            .filter {it.parameters.length == callArgsList.size()}
                            .filter {
                                for (def i1 = 0; i1 < callArgsList.size(); i1 ++) {
                                    if (!isArgumentTypeAndParamSuitable(callArgsTypes[i1], it.parameters[i1])) {
                                        return false
                                    }
                                }

                                return true
                            }
                            .findAny().orNull()

                    if (targetMethod == null) {
                        def parameterMatchingExtensions = availableExtensions.stream()
                            // Идем от методов с наибольшим кол-вом аргументов
                            .reverseSortedBy {it.method.parameters.length}
                            // Забираем все расширения, у которых кол-во параметров без учета первого self-параметра подходит под вызов целевого метода
                            .filter {
                                it.method.parameters.stream()
                                    .filter {p -> !p.hasInitialExpression()}
                                    .count().mapTo {count -> count - 1} <= callArgsList.size()
                            }
                            .filter {
                                it.method.parameters.length - 1 >= callArgsList.size()
                            }

                            // Проверяем, что переданные в метод аргументы подходят под метод-расширение
                            .filter {
                                for (def i1 = 0; i1 < callArgsList.size(); i1 ++)
                                    if (!isArgumentTypeAndParamSuitable(callArgsTypes[i1], it.method.parameters[i1 + 1]))
                                    {
                                        return false
                                    }

                                return true
                            }

                            .toList()
                            .mapTo {

                                // Если подошло больше, чем одно расширение - это ошибка. Значит криво настроены сигнатуры
                                if (it.size() > 1)
                                    addError("More than one extensions found for method!", callExpression)
                                else if (it.isNotNullOrEmpty())
                                    return it.first()

                                return (ExtensionMethodInfo) null
                            }

                        // Если нашли расширение, то подменяем вызов несуществующего метода на вызов расширения
                        if (parameterMatchingExtensions != null) {
                            def newCallArgs = new ArrayList<Expression>(callArgsList)
                            if (isCallStatic) {
                                newCallArgs.add(0, nullX())
                            } else {
                                newCallArgs.add(0, receiverExpression as Expression)
                            }

                            return new StaticMethodCallExpression(parameterMatchingExtensions.method.declaringClass, parameterMatchingExtensions.method.name, args(newCallArgs))
                        }
                    }
                }
            }
        }

        return callExpression
    }

    boolean isArgumentTypeAndParamSuitable(ClassNode argType, Parameter param) {
        if (argType == null && ClassHelper.isPrimitiveType(param.type)) {
            return false
        }

        if (argType != null && !isNodeAssignable(argType, param.type)) {
            return false
        }

        return true
    }

    //TODO: Не будет работать с массивами! Доработать!
    List<ExtensionMethodInfo> getOwnerMatchingStaticExtensions(ClassNode cn) {
        return staticExtensions.stream()
            .filter { isNodeAssignable(cn, it.targetType)}
            .toList()
    }

    List<ExtensionMethodInfo> getOwnerMatchingInstanceExtensions(ClassNode cn) {
        return instanceExtensions.stream()
            .filter { isNodeAssignable(cn, it.targetType)}
            .toList()
    }

    static boolean isNodeAssignable(ClassNode cn, ClassNode assignableCn) {
        def isFirstPrimitive = ClassHelper.isPrimitiveType(cn)
        def isSecondPrimitive = ClassHelper.isPrimitiveType(assignableCn)
        if (isFirstPrimitive != isSecondPrimitive) {
            def object1Type = isFirstPrimitive ? ClassHelper.getWrapper(cn) : cn
            def object2Type = isSecondPrimitive ? ClassHelper.getWrapper(assignableCn) : assignableCn

            return isNodeAssignable(object1Type, object2Type)
        }

        def parent = cn
        while (parent != null) {
            if (parent.name == assignableCn.name) {
                return true
            }

            if (parent?.interfaces?.any {it.name == assignableCn.name}) {
                return true
            }

            parent = parent.superClass
        }

        return false
    }

    @Override
    protected SourceUnit getSourceUnit() {
        return sourceUnit
    }

    // Понятия не имею почему, но если добавлять этот AST в груви через конфиг для компилятора у градла, то овнеры у статических вызовов оборачиваются в java.lang.Class
    protected ClassNode unpackStaticClassReciever(ClassNode cn) {
        if (cn == null || cn.name != "java.lang.Class") {
            return cn
        }

        if (cn?.genericsTypes?.length == 1) {
            return cn.genericsTypes[0].type
        }

        return cn
    }
}
